/******************************************************************************
 * 
 *  textable - TeX tables generator. GNUPlot's data to TeX's tabular.
 *  
 ******************************************************************************
 *
 *  Copyright (C) 2003 Josef Kufner
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 ******************************************************************************
 *
 * TODO
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "main.h"
#include "debug.h"
#include "lists.h"
#include "parser.h"
#include "calc.h"
#include "generator.h"

int verbose = 0, debug = 0, ignore_error = 0;
char *prg_name = NULL;
char *in_file = NULL, *out_file = NULL, *out_plain_file = NULL;

/******************************************************************************
 *	malloc_c -- enhanced malloc (+ error check)
 *****************************************************************************/
void *malloc_c (const size_t size) {
	void *mem;

	mem = malloc(size);
	if (!mem) {
		PRINT_ERR_PRG("Error while alocating memory: %s",
				strerror(errno));
		exit(2);
	}
	return mem;
}



/******************************************************************************
 *	on_Exit -- call this when exiting
 *****************************************************************************/
void on_textable_exit (void) {
	unsigned int i;

	PRINT_DEBUG_EMPTY_LINE();
	PRINT_DEBUG_PRG("%s", "Exiting...");

	/* zavirani souboru */
	parser_input_end();
	gen_latex_end();
	gen_plain_end();

	
	/* uvolneni pameti */
	if (in_file)
		free(in_file);
	if (out_file)
		free(out_file);
	if (out_plain_file)
		free(out_plain_file);

	
	/* Likvidace seznamu */	
	for (i = 0; i < MAX_LIST; i++)
		destroy_list(i);

	/* Say Bye */
	PRINT_VERBOSE_PRG("%s", "Bye.");
	
	if (prg_name)
		free(prg_name);
}



/******************************************************************************
 *	main
 *****************************************************************************/
int main (int argc, char *argv[]) {
	int opt;
	char *line;
	LIST_ITEM *item, *format_item, *ex_item;
	int is_first_data = 1;
	
	/* init global variables */
	prg_name = (char *) malloc_c((strlen(argv[0]) + 1) * sizeof(char));
	strcpy(prg_name, argv[0]);

	/* init atexit */
	atexit(on_textable_exit);
	
	/* Args */
	while ((opt = getopt(argc, argv, "i:o:p:vdDe")) != -1) {
		switch(opt) {
			case 'i':
				if (in_file)
					free(in_file);

				in_file = (char *) malloc_c((strlen(optarg) + 1) * sizeof(char));
				strcpy(in_file, optarg);
				break;
			case 'o':
				if (out_file)
					free(out_file);
				
				out_file = (char *) malloc_c((strlen(optarg) + 1) * sizeof(char));
				strcpy(out_file, optarg);
				break;
			case 'p':
				if (out_plain_file)
					free(out_plain_file);

				out_plain_file = (char *) malloc_c((strlen(optarg) + 1) * sizeof(char));
				strcpy(out_plain_file, optarg);
				break;
			case 'v':
				verbose = 1;
				break;
			case 'd':
				debug = 1;
				break;
			case 'D':
				debug = 2;
				break;
			case 'e':
				ignore_error = 1;
				break;
			case 'h':
			case '?':
				fprintf(stderr, "usage: %s [-i <input_file>] [-o <output_LaTeX_file>] "
						"[-p <output_plain_text_file>] [-v]\n",
						prg_name);
				return EXIT_ERROR;
				break;
		}
	}
	
	/* Verbose info */
	PRINT_VERBOSE_PRG("input:        \"%s\"", in_file ? in_file : "<stdin>");
	PRINT_VERBOSE_PRG("LaTeX output: \"%s\"", out_file ? out_file : "<stdout>");
	PRINT_VERBOSE_PRG("plain output: \"%s\"", out_plain_file ? out_plain_file : "<none>");

	/* prepare input */
	parser_input_begin(in_file);

	/* prepare generated files */
	gen_latex_begin(out_file);
	gen_plain_begin(out_plain_file);

	/* prepare lists */
	create_list(ADD_LIST);
	create_list(VAR_LIST);
	create_list(HEAD_LIST);
	create_list(FORMAT_LIST);
	
	/* Processing input **************************************************/
	
	while ((line = read_line()) != NULL) {
		
		if (parse_line(line)) {
			/* line parsed */

			/* Process Data list -- if there are no data on the line, data list is empty */
			if (!is_list_empty(DATA_LIST)) {
				
				PRINT_DEBUG_PRG_L2("%s", "Data_List is NOT empty.");
				/* if we have first data, write header */
				if (is_first_data) {
					gen_latex_head();
				}

				/* process data list */
				rewind_list(DATA_LIST);
				push_list_pos(FORMAT_LIST);
				rewind_list(FORMAT_LIST);
				while ((item = get_item(DATA_LIST)) != NULL) {
					format_item = get_item(FORMAT_LIST);
					strcpy(item->val, calc(subst(item->expr, DATA_LIST),
									format_item ? format_item->format_mode : 0,
									format_item ? format_item->format_digits : 2));
				}
				
				pop_list_pos(FORMAT_LIST);

				
				/* if example_save flag is set, copy data to example list */

				if (example_save) {
					PRINT_DEBUG_PRG_L2("%s", "Copying data_list to example_list...");

					rewind_list(DATA_LIST);
					while ((item = get_item(DATA_LIST)) != NULL)
						PRINT_DEBUG_PRG("Item: \"%s\" = \"%s\"", item->name, item->expr);
					
					clear_list(EXAMPLE_LIST);
				
					push_list_pos(DATA_LIST);
					rewind_list(DATA_LIST);
					
					while ((item = get_item(DATA_LIST)) != NULL) {
						ex_item = create_item(EXAMPLE_LIST);
						memcpy(ex_item, item, sizeof(LIST_ITEM));
						PRINT_DEBUG_PRG_L2("Item: \"%s\" = \"%s\"", item->name, item->expr);
					}

					pop_list_pos(DATA_LIST);

					gen_set_example_line();

					PRINT_DEBUG_PRG_L2("%s", "Example:");

					rewind_list(EXAMPLE_LIST);
					while ((item = get_item(EXAMPLE_LIST)) != NULL)
						PRINT_DEBUG_PRG_L2("Item: \"%s\" = \"%s\"", item->name, item->expr);
					getc(stdin);
				
					PRINT_DEBUG_EMPTY_LINE();

					example_save = 0;
				}	
				
				
				/* put data to output */
				gen_latex_data();
				gen_plain_data();
				
				/* ok, we have some data, next data can't be first */
				is_first_data = 0;
			} else {
				/* if data list is empty, there is some command on the line,
				 * but it was processed by parse_line() */
				PRINT_DEBUG_PRG_L2("%s", "Data_List is empty.");
			}
			
		} else {
			/* line ignored */
		}

		clear_list(DATA_LIST);
	}

	/* TODO
	 * priklad vypoctu */

	/* end of generated files */
	/* this will be done by on_textable_exit */
	
	return EXIT_OK;
}
