/*
 * parser.c
 *
 * Josef Kufner <jkufner@seznam.cz>
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "main.h"
#include "debug.h"
#include "parser.h"
#include "lists.h"
#include "generator.h"

#define BUF_STEP	128

static char *return_data;
static int hd_no = 0, add_no = 0;
int line_no = 0;

int example_save = 0;

FILE *f_src;

/*****************************************************************************/

static char *skip_space (char *p_line) {

	while (*p_line != '\0' && isspace(*p_line)) {
		p_line += sizeof(char);
	}

	return p_line;
}


/*****************************************************************************/

static void cmdfn_format(char *p_line) {

	PRINT_DEBUG("Table format: \"%s\"", p_line);
	gen_set_format(p_line);
}

static void cmdfn_hd(char *p_line) {
	LIST_ITEM *item;

	hd_no++;

	item = create_item(HEAD_LIST);

	sprintf(item->name, "%d", hd_no);
	strcpy(item->expr, p_line);

	PRINT_DEBUG("Heading: $%s = \"%s\"", item->name, item->val);
}

static void cmdfn_hend(char *p_line) {
	LIST_ITEM *item;

	hd_no = 0;
	/* Do NOT destroy head_list, bacause we need it. */

	PRINT_DEBUG("%s", "Heading end.");

	rewind_list(HEAD_LIST);
	/*
	while ((item = get_item(HEAD_LIST)) != NULL) {
		PRINT_DEBUG("\t$%s = \"%s\"", item->name, item->expr);
	}
	*/
}


static void cmdfn_add(char *p_line) {
	LIST_ITEM *item;

	add_no++;

	item = create_item(ADD_LIST);
	sprintf(item->name, "_%d", add_no);
	strcpy(item->expr, p_line);
	
	PRINT_VERBOSE("Additional column: $%s = \"%s\"", item->name, item->expr);
}


static void cmdfn_var(char *p_line) {
	LIST_ITEM *item;
	int i;

	item = create_item(VAR_LIST);

	/* if there is leading '$' skip it */
	if (*p_line == '$')
		p_line += sizeof(char);

	i = 0;
	while(i < MAX_LEN - 1 && *p_line != '\0' && (isalnum(*p_line) || *p_line == '_')) {
		item->name[i++] = *p_line;
		item->name[i] = '\0';
		p_line += sizeof(char);
	}

	p_line = skip_space(p_line);

	if (*p_line == '=') {
		p_line += sizeof(char);
		p_line = skip_space(p_line);
	}
	
	strcpy(item->expr, p_line);
	
	PRINT_VERBOSE("Variable: $%s = \"%s\"", item->name, item->expr);
}

static void cmdfn_colformat(char *p_line) {
	LIST_ITEM *item;

	item = create_item(FORMAT_LIST);

	if (*p_line == 'n') {
		item->format_mode = NORMAL;
		p_line += sizeof(char);
	} else if (*p_line == 's' || *p_line == 'e') {
		item->format_mode = SCI;
		p_line += sizeof(char);
	}

	if (isdigit(*p_line)) {
		sscanf(p_line, "%d", &(item->format_digits));
	}
	
}

static void cmdfn_raw (char *p_line) {
	gen_latex_raw(p_line);
}

static void cmdfn_exline (char *p_line) {

	example_save = 1;
}

static void cmdfn_example(char *p_line) {
	verbose = 1; debug = 2;
	PRINT_VERBOSE_PRG("%s", "Generating examples...");
	gen_latex_example();
}

/*****************************************************************************/


int parser_input_begin(char *file_name) {
	
	if (file_name) {
		f_src = fopen(file_name, "rt");
		if (f_src == NULL) {
			PRINT_ERR_PRG("Error while opening input file: %s", strerror(errno));
			exit(EXIT_ERROR);
		}
		return 1;
	} else {
		f_src = stdin;
		return 1;
	}
	return 0;
}
int parser_input_end(void) {

	if (f_src) {
		if (fclose(f_src))
			return 0;
		return 1;
	}
	return 0;
}


/*
 * read_line -- nacte jeden radek ze souboru
 */
char *read_line (void) {
	char *buf = NULL, ch;
	static char *line_buf = NULL;
	int i;
	
	/* Nejdrive inkrementujem pocitadlo radku */
	line_no++;

	PRINT_DEBUG_EMPTY_LINE();
	PRINT_DEBUG_PRG_L2("Reading line %d...", line_no);

	/* Free old string */
	if (line_buf) {
		free(line_buf);
		line_buf = NULL;
	}
	
	/* initial buf size */
	line_buf = (char *) malloc_c(BUF_STEP * sizeof(char));
	line_buf[0] = '\0';
	
	if (f_src == NULL || feof(f_src)) {
		PRINT_DEBUG_PRG("EOF on line %d.", line_no);
		return NULL;
	}
	
	/* reading */
	for (i = 0; !feof(f_src) && (ch = getc(f_src)) != '\n'; i++) {

		/* replace \t with space - only for debug */
		ch = ch == '\t' ? ' ' : ch;
		
		/* enlarge buffer */
		if (i % BUF_STEP == BUF_STEP - 1) {
			PRINT_DEBUG_PRG("Line %d too long -- enlarging input buffer to %d bytes.",
						line_no, (i + BUF_STEP + 1) * sizeof(char));
			/* backup */
			buf = (char *) malloc_c(i * sizeof(char) + 1);
			memcpy(buf, line_buf, i * sizeof(char));
			free(line_buf);
			
			/* copy back */
			line_buf = (char *) malloc_c((i + BUF_STEP) * sizeof(char) + 1);
			memcpy(line_buf, buf, i * sizeof(char));
			free(buf);
		}

		/* add char to buffer */
		line_buf[i] = ch;
		line_buf[i+1] = '\0';
	}
	
	/* return value */
	return line_buf;
}

/* parse_line -- zpracuje retezec a naplni odpovidajici seznamy (seznamy musi byt pripraveny)
 *
 * ret val:
 * 	- pokud se jednalo o komentar, prazdnou radku nebo se to nepovedlo, vraci 0
 * 	- pokud probehlo vse ok, vraci 1
 */
int parse_line(char *line) {
	int col_no = 0;
	char *p_line = NULL;
	char cmd[32];
	int i;
	LIST_ITEM *item, *orig_item;
	
	const struct {
		char *name;
		void (*fn)(char *);
	} command[] = {
		{ "FORMAT",	cmdfn_format	},
		{ "HD",		cmdfn_hd	},
		{ "HEND",	cmdfn_hend	},
		{ "ADD",	cmdfn_add	},
		{ "VAR",	cmdfn_var	},
		{ "COLFORMAT",	cmdfn_colformat	},
		{ "RAW",	cmdfn_raw	},
		{ "EXLINE",	cmdfn_exline	},
		{ "EXAMPLE",	cmdfn_example	},
		{ NULL,		NULL		}
	};

	
	PRINT_DEBUG_PRG_L2("Parsing line %d...", line_no);

	p_line = skip_space(line);

	if (feof(f_src))
		return 0;
	
	if (p_line[0] == '#') {
		if (p_line[1] == '>') {
			/* Command ****************************/

			p_line += 2 * sizeof(char);

			/* copy command */
			for(i = 0; *p_line != '\0' && isalpha(*p_line) && i < 32 - 1; i++) {
				cmd[i] = *p_line;
				cmd[i+1] = '\0';
				p_line += sizeof(char);
			}
			
			p_line = skip_space(p_line);

			for (i = 0; command[i].name != NULL; i++) {
				if (strcmp(command[i].name, cmd) == 0) {
					(*command[i].fn)(p_line);
					break;
				}
			}

			if (command[i].name == NULL) {
				PRINT_ERR_FILE("Unknown command: %s%s", cmd, ignore_error ? " -- ignoring" : "");
				if (!ignore_error)
					exit(EXIT_ERROR);
			}
		
		} else {
			/* Comment ****************************/
			PRINT_DEBUG_L2("Comment: %s", p_line);
		}
	} else if(*p_line != '\0') {
		PRINT_DEBUG_L2("%s", "Data detected...");
		
		clear_list(DATA_LIST);
	
		if (example_save)
			clear_list(EXAMPLE_LIST);
		
		col_no = 0;

		/* Read data from input */
		while (*p_line != '\0') {
			
			p_line = skip_space(p_line);

			if (*p_line == '\0' || *p_line == '#')
				break;

			col_no++;
			item = create_item(DATA_LIST);
			
			sprintf(item->name, "%d", col_no);
			
			i = 0;
			while(*p_line != '\0' && *p_line != '#' && !isspace(*p_line)) {
				item->expr[i++] = *p_line;
				item->expr[i] = '\0';
				p_line += sizeof(char);
			}

			PRINT_DEBUG("Data: $%s = \"%s\"", item->name, item->expr);
		}

		/* Generate additional data */
		rewind_list(ADD_LIST);
		while((orig_item = get_item(ADD_LIST))) {
			
			col_no++;
			item = create_item(DATA_LIST);
			sprintf(item->name, "%d", col_no);
			strcpy(item->expr, orig_item->expr);

			PRINT_DEBUG("Additional Data: $%s = \"%s\"", item->name, item->expr);
		}


		
	} else {
		PRINT_DEBUG_L2("%s", "Empty line.");
	}

	return 1;
}

